llvm-project/clang/test/SemaCUDA/constexpr-variables.cu
Yaxun (Sam) Liu 049d860707 [CUDA][HIP] Fix constexpr variables for C++17
constexpr variables are compile time constants and implicitly const, therefore
they are safe to emit on both device and host side. Besides, in many cases
they are intended for both device and host, therefore it makes sense
to emit them on both device and host sides if necessary.

In most cases constexpr variables are used as rvalue and the variables
themselves do not need to be emitted. However if their address is taken,
then they need to be emitted.

For C++14, clang is able to handle that since clang emits them with
available_externally linkage together with the initializer.

However for C++17, the constexpr static data member of a class or template class
become inline variables implicitly. Therefore they become definitions with
linkonce_odr or weak_odr linkages. As such, they can not have available_externally
linkage.

This patch fixes that by adding implicit constant attribute to
file scope constexpr variables and constexpr static data members
in device compilation.

Differential Revision: https://reviews.llvm.org/D79237
2020-06-03 21:56:52 -04:00

81 lines
2.6 KiB
Plaintext

// RUN: %clang_cc1 -std=c++14 %s -emit-llvm -o - -triple nvptx64-nvidia-cuda \
// RUN: -fcuda-is-device -verify -fsyntax-only
// RUN: %clang_cc1 -std=c++17 %s -emit-llvm -o - -triple nvptx64-nvidia-cuda \
// RUN: -fcuda-is-device -verify -fsyntax-only
// RUN: %clang_cc1 -std=c++14 %s -emit-llvm -o - \
// RUN: -triple x86_64-unknown-linux-gnu -verify -fsyntax-only
// RUN: %clang_cc1 -std=c++17 %s -emit-llvm -o - \
// RUN: -triple x86_64-unknown-linux-gnu -verify -fsyntax-only
#include "Inputs/cuda.h"
template<typename T>
__host__ __device__ void foo(const T **a) {
// expected-note@-1 {{declared here}}
static const T b = sizeof(a);
static constexpr T c = sizeof(a);
const T d = sizeof(a);
constexpr T e = sizeof(a);
constexpr T f = **a;
// expected-error@-1 {{constexpr variable 'f' must be initialized by a constant expression}}
// expected-note@-2 {{read of non-constexpr variable 'a' is not allowed in a constant expression}}
a[0] = &b;
a[1] = &c;
a[2] = &d;
a[3] = &e;
}
__device__ void device_fun(const int **a) {
// expected-note@-1 {{declared here}}
constexpr int b = sizeof(a);
static constexpr int c = sizeof(a);
constexpr int d = **a;
// expected-error@-1 {{constexpr variable 'd' must be initialized by a constant expression}}
// expected-note@-2 {{read of non-constexpr variable 'a' is not allowed in a constant expression}}
a[0] = &b;
a[1] = &c;
foo(a);
// expected-note@-1 {{in instantiation of function template specialization 'foo<int>' requested here}}
}
void host_fun(const int **a) {
// expected-note@-1 {{declared here}}
constexpr int b = sizeof(a);
static constexpr int c = sizeof(a);
constexpr int d = **a;
// expected-error@-1 {{constexpr variable 'd' must be initialized by a constant expression}}
// expected-note@-2 {{read of non-constexpr variable 'a' is not allowed in a constant expression}}
a[0] = &b;
a[1] = &c;
foo(a);
}
__host__ __device__ void host_device_fun(const int **a) {
// expected-note@-1 {{declared here}}
constexpr int b = sizeof(a);
static constexpr int c = sizeof(a);
constexpr int d = **a;
// expected-error@-1 {{constexpr variable 'd' must be initialized by a constant expression}}
// expected-note@-2 {{read of non-constexpr variable 'a' is not allowed in a constant expression}}
a[0] = &b;
a[1] = &c;
foo(a);
}
template <class T>
struct A {
explicit A() = default;
};
template <class T>
constexpr A<T> a{};
struct B {
static constexpr bool value = true;
};
template<typename T>
struct C {
static constexpr bool value = T::value;
};
__constant__ const bool &x = C<B>::value;